import random
import re

import autopep8
from PyQt5 import QtGui
from PyQt5.QtCore import pyqtSignal, Qt, QTimer, QMimeData
from PyQt5.QtGui import QFont, QFontMetrics, QColor, QPainter, QTextCursor
from PyQt5.QtWidgets import QTextEdit, QHBoxLayout, QScrollBar, QApplication

from spiceditor.line_number_text_edit import LineNumberTextEdit
from spiceditor.magic_scrollbar import MagicScrollBar


class SpiceMagicEditor(QTextEdit):
    ctrl_enter = pyqtSignal()
    info = pyqtSignal(str, int, int)

    def __init__(self, highlighter=None, font_size=18):
        super().__init__()
        self.highlighter = highlighter
        self.suggestion = None
        self.candidates = []
        self.count = 0
        self.mode = 0
        self.code = ""
        self.delay = 0.01
        self.autocomplete_words = []

        self.line_number_area = LineNumberTextEdit(self)
        self.line_number_area.verticalScrollBar().valueChanged.connect(self.verticalScrollBar().setValue)
        self.line_number_area.document().setDocumentMargin(0)

        self.setContentsMargins(0, 0, 0, 0)
        self.document().setDocumentMargin(0)
        self.setViewportMargins(60, 0, 0, 0)
        self.setLineWrapMode(QTextEdit.NoWrap)
        self.setPlaceholderText("Write Python code here...")

        self.setHorizontalScrollBar(MagicScrollBar())
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOn)

        self.textChanged.connect(self.text_changed)
        self.verticalScrollBar().valueChanged.connect(self.line_number_area.verticalScrollBar().setValue)
        self.horizontalScrollBar().rangeChanged.connect(self.text_changed)

        self.cursorPositionChanged.connect(
            lambda: self.line_number_area.highlight_line(self.textCursor().blockNumber()))

        if self.highlighter:
            self.highlighter.setDocument(self.document())

        self.set_font_size(font_size)

    def insertFromMimeData(self, source: QMimeData):
        """Override paste behavior to insert plain text only."""
        if source.hasText():
            self.insertPlainText(source.text())

    def set_font_size(self, font_size):
        font = QFont("Courier New")
        # font.setStyleHint(QFont.TypeWriter)
        font.setPixelSize(font_size)
        self.setFont(font)
        self.line_number_area.setFont(font)
        three_numbers_width = QFontMetrics(font).width("000")
        self.line_number_area.setFixedWidth(int(three_numbers_width + 12))
        self.setViewportMargins(three_numbers_width + 15, 0, 0, 0)

    def show_code(self):
        self.show_all_code()

    def set_dark_mode(self, dark):
        self.line_number_area.set_dark_mode(dark)
        self.highlighter.set_dark_mode(dark)
        self.highlighter.setDocument(self.document())

    def set_text(self, text):
        self.setPlainText(text)
        self.text_changed()

    def text_changed(self):
        text = self.toPlainText()
        lines = text.split("\n")
        v1 = self.line_number_area.verticalScrollBar().value()
        self.line_number_area.blockSignals(True)
        self.line_number_area.clear()
        text = str()
        for i in range(len(lines)):
            text += "{:3d}\n".format(i + 1)

        self.line_number_area.setPlainText(text)
        self.line_number_area.setFixedHeight(self.height())
        self.line_number_area.verticalScrollBar().setMaximum(self.verticalScrollBar().maximum())
        self.line_number_area.verticalScrollBar().setValue(v1)
        self.line_number_area.blockSignals(False)

    def resizeEvent(self, a0) -> None:
        self.text_changed()
        self.blockSignals(True)
        super().resizeEvent(a0)
        self.blockSignals(False)

    def format_code(self):
        self.format_code()

    def get_text(self):
        return self.toPlainText()

    def clear(self):
        self.set_code("")
        self.setPlainText("")
        self.count = 0
        self.set_mode(0)

    def get_code(self):
        return self.code

    def get_remaining_chars(self):
        diff = len(self.get_code()) - self.count
        return diff

    def set_delay(self, delay):
        self.delay = delay

    def append_autocomplete(self, words, clear=False):
        if clear:
            self.autocomplete_words.clear()
        self.autocomplete_words += words if words else ""

    def set_code(self, code):
        self.setText("")
        self.count = 0
        self.code = code
        self.set_mode(1)
        self.setFocus()

    def set_mode(self, mode):
        self.mode = mode
        self.setCursorWidth(3 if self.mode == 1 else 1)
        # self.setReadOnly(self.mode == 1)
        self.update()

    def on_return_key(self, e):
        return False

    def complete_line(self, sleep=True):
        self.info.emit(self.get_next_line(), self.get_remaining_chars(), 20)
        if self.count < len(self.code):
            # self.setText(self.toPlainText() + self.code[self.count])
            self.insertPlainText(self.code[self.count])
            self.moveCursor(QtGui.QTextCursor.End)
            self.count += 1

            if self.code[self.count - 1] == "\n":
                # if next line is empty continue
                # and show that line too
                if len(self.get_rest_of_line()) > 0:
                    return True

            QApplication.processEvents()
            delay = int(self.delay) + random.randint(0, int(self.delay))
            QTimer.singleShot(delay, self.complete_line)

    def get_rest_of_line(self):
        count = self.count
        text = ""
        while count < len(self.code):
            text += self.code[count]
            count += 1
            if self.code[count - 1] == "\n":
                return text[1:]
        return ""

    def get_spaces(self, line):
        spaces = 0
        for c in line:
            if c == " ":
                spaces += 1
            else:
                break
        return spaces

    def get_current_line(self):
        cursor = self.textCursor()
        cursor.movePosition(QtGui.QTextCursor.StartOfLine)
        cursor.movePosition(QtGui.QTextCursor.EndOfLine, QtGui.QTextCursor.KeepAnchor)
        current_line = cursor.selectedText()
        return current_line

    def append_next_char(self):
        self.count += 1
        self.setText(self.code[:self.count])
        self.moveCursor(QtGui.QTextCursor.End)

    def show_all_code(self):
        self.setText(self.code)
        self.set_mode(0)
        self.moveCursor(QtGui.QTextCursor.End)

    def get_current_line_text(self):
        # Get the QTextCursor
        cursor = self.textCursor()

        # Move the cursor to the start and end of the current line
        cursor.select(cursor.LineUnderCursor)

        # Get the selected text
        return cursor.selectedText()

    def keyPressEvent(self, e: QtGui.QKeyEvent) -> None:
        self.setFocusPolicy(Qt.StrongFocus)

        if e.key() == Qt.Key_Escape:
            self.set_mode(0 if self.mode == 1 else 0)

        if self.mode == 1:

            if e.key() == Qt.Key_Down:
                self.show_all_code()
            elif e.key() == Qt.Key_Control:
                return
            elif e.key() == Qt.Key_End:
                while self.complete_line(False):
                    pass
                self.set_mode(0)
            elif e.key() == Qt.Key_Tab:
                if e.modifiers() == Qt.ControlModifier:
                    while self.complete_line():
                        pass
                    self.set_mode(0)
                else:
                    self.complete_line()
            elif self.count < len(self.code):
                self.info.emit(self.get_rest_of_line(), self.get_remaining_chars(), 1000)
                self.append_next_char()
            elif e.key() == Qt.Key_Return:
                self.set_mode(0)
                super().keyPressEvent(e)
            elif e.key() == Qt.Key_Backspace:
                self.set_mode(0)
            else:
                self.setText(self.toPlainText() + "\n")
                self.moveCursor(QtGui.QTextCursor.End)

        elif self.mode == 0:

            if e.key() == Qt.Key_Tab:
                self.tab_pressed()
            elif e.key() == Qt.Key_Backspace:
                self.suggestion = None
                if self.get_current_line_text().endswith("    "):
                    for i in range(4):
                        self.textCursor().deletePreviousChar()
                else:
                    super().keyPressEvent(e)
            elif e.key() == Qt.Key_Return:
                self.suggestion = None
                if e.modifiers() == Qt.ControlModifier:
                    self.ctrl_enter.emit()
                elif self.on_return_key(e):
                    pass
                else:
                    super().keyPressEvent(e)
            else:
                self.suggestion = None
                super().keyPressEvent(e)
        self.cursorPositionChanged.emit()

    def indent_selected(self):
        cursor = self.textCursor()

        if not cursor.hasSelection():
            return  # No selection, do nothing

        # Get selected text (preserve newlines)
        selected_text = cursor.selection().toPlainText()

        # Add 4 spaces to each line
        indented_text = "\n".join("    " + line for line in selected_text.splitlines())

        # Replace selected text with indented version
        cursor.beginEditBlock()  # Start undo-able action
        cursor.insertText(indented_text)
        cursor.endEditBlock()  # End undo-able action

    def get_text_before_cursor(self):
        cursor = self.textCursor()
        cursor.movePosition(QTextCursor.StartOfBlock, QTextCursor.KeepAnchor)  # Selecciona desde el inicio de la línea
        return cursor.selectedText()

    def tab_pressed(self):

        if self.textCursor().hasSelection():
            self.indent_selected()
            return

        # Let see if we have some autocomplete candidates
        if self.suggestion is None:
            current_words = re.split(r'\W+', self.toPlainText())
            text_before_cursor = self.get_text_before_cursor()
            words_before_cursos = re.split(r"[+\-*/= ]", text_before_cursor)
            self.candidates = []
            if words_before_cursos[-1] != "":
                word_set = list(set(current_words + self.highlighter.get_keywords() + self.autocomplete_words))
                self.candidates = [word for word in word_set if word.startswith(words_before_cursos[-1])]
                if words_before_cursos[-1] in self.candidates:
                    self.candidates.remove(words_before_cursos[-1])
                    self.candidates.append(words_before_cursos[-1])
                self.suggestion = words_before_cursos[-1]

        if len(self.candidates) > 1:
            # Remove the current suggestion
            for _ in range(len(self.suggestion)):
                self.textCursor().deletePreviousChar()
            self.candidates.append(self.suggestion)
            self.suggestion = self.candidates.pop(0)
            self.insertPlainText(self.suggestion)
        else:
            self.insertPlainText("    ")



        # if len(self.candidates) > 0:

        # current_line = self.get_current_line_text()
        #
        # # it was just a tab
        # if len(current_line) == 0 or current_line.endswith("    "):
        #     self.insertPlainText("    ")
        #     self.moveCursor(QtGui.QTextCursor.End)
        #     self.suggestion = None
        #     return
        #
        # # it was just a tab
        # if len(current_line) > 0 and current_line[-1] in " (:)":
        #     self.suggestion = None
        #     return
        #
        # if self.suggestion is None:
        #
        #     if len(self.candidates) == 0:
        #         self.insertPlainText("    ")
        #         return
        #     elif len(self.candidates) == 1:
        #         self.insertPlainText(self.candidates[0][len(unfinished_word):])
        #         self.moveCursor(QtGui.QTextCursor.End)
        #         return
        #     else:
        #         self.suggestion = unfinished_word
        #
        # for _ in range(len(self.suggestion)):
        #     self.textCursor().deletePreviousChar()
        # self.candidates.append(self.suggestion)
        # self.suggestion = self.candidates.pop(0)
        # self.insertPlainText(self.suggestion)

    def get_next_line(self):
        count = self.count
        remaining = self.code[count:]
        remaining = remaining.split("\n")
        remaining = remaining[1:]
        remaining = [x for x in remaining if x.strip()]
        if remaining:
            return remaining[0]


class PascalEditor(SpiceMagicEditor):
    pass


class PythonEditor(SpiceMagicEditor):
    def format_code(self):
        code = self.toPlainText()
        if not code.endswith("\n"):
            code += "\n"
        self.setPlainText(autopep8.fix_code(code))
        self.moveCursor(QtGui.QTextCursor.End)

    def on_return_key(self, e):
        current_line = self.get_current_line()
        spaces = self.get_spaces(current_line)
        if current_line.endswith(":"):
            if self.textCursor().positionInBlock() == len(current_line):
                self.insertPlainText("\n" + " " * (spaces + 4))
                return True
            return False

        elif current_line.startswith(" "):
            if current_line.strip():
                self.insertPlainText("\n" + " " * spaces)
            else:
                self.insertPlainText("\n" + " " * (max(spaces - 4, 0)))
            return True
        return False



import os
import platform
import subprocess

import fitz  # PyMuPDF
import sys

import fitz  # PyMuPDF
from PyQt5.QtCore import Qt, QTimer, pyqtSignal, QLine, QLineF, QRectF, QPointF
from PyQt5.QtGui import QPixmap, QImage, QFont, QPainter, QColor, QCursor, QIcon, QTransform, QPen
from PyQt5.QtWidgets import QMainWindow, QLabel, QSizePolicy, QApplication, QVBoxLayout, QWidget, QPushButton, \
    QShortcut, QInputDialog, QTabBar, QToolBar, QHBoxLayout, QComboBox, QGraphicsView, QGraphicsScene, \
    QGraphicsPixmapItem, QGraphicsItem, QGraphicsEllipseItem, \
    QGraphicsRectItem, QGraphicsProxyWidget, QGraphicsLineItem
from pymupdf import Rect
from scipy.signal import savgol_filter

from spiceditor.utils import create_cursor_image


class Eraser(QGraphicsEllipseItem):
    pass


def smooth_with_savgol(points, window_size=10, poly_order=2):
    if len(points) < window_size:
        return points
    x = [p[0] for p in points]
    y = [p[1] for p in points]
    smoothed_x = savgol_filter(x, window_size, poly_order)
    smoothed_y = savgol_filter(y, window_size, poly_order)
    return list(zip(smoothed_x, smoothed_y))


class GraphicsScene(QGraphicsScene):
    NONE = 0
    POINTER = 1
    WRITING = 2
    ERASING = 3
    RECTANGLES = 4
    ELLIPSES = 5

    navigate = pyqtSignal(int)

    def keyPressEvent(self, event):
        super().keyPressEvent(event)
        if event.key() == Qt.Key_E:
            self.status = GraphicsScene.ERASING
        elif event.key() == Qt.Key_W:
            self.status = GraphicsScene.WRITING
        elif event.key() in [Qt.Key_Right, Qt.Key_Down]:
            self.navigate.emit(1)
        elif event.key() in [Qt.Key_Left, Qt.Key_Up]:
            self.navigate.emit(-1)

    def __init__(self):
        super().__init__()
        self.handwriting = []
        self.pixmap = QGraphicsPixmapItem()
        self.pixmap.setTransformationMode(Qt.SmoothTransformation)  # Enable smooth transformation for the pixmap item

        self.addItem(self.pixmap)
        self.image = None
        self.start = None
        self.page = 0
        self.status = GraphicsScene.NONE

        self.gum = Eraser(0, 0, 100, 100, self.pixmap)
        self.gum.setBrush(QColor(255, 255, 255))
        self.gum.setPen(QPen(QColor(0, 0, 0), 2))
        self.gum.setVisible(False)
        self.drawings = {}
        self.rectangle = None
        self.color = QPen(Qt.black, 2)
        # make the ellipse movable
        # self.gum.setFlag(QGraphicsItem.ItemIsMovable)

    def set_image(self, image, page):
        current = self.drawings.get(self.page, [])
        for item in current:
            self.removeItem(item)

        self.image = image
        self.page = page
        # self.resize_image(size)
        self.pixmap.setPixmap(self.image)

        for item in self.drawings.get(self.page, []):
            self.addItem(item)

    def mousePressEvent(self, event):
        event.accept()
        super().mousePressEvent(event)
        if event.button() == Qt.RightButton:
            return

        self.start = event.scenePos()

        # get objects at the click position
        items = self.items(self.start)
        for item in items:
            if type(item) not in [QGraphicsPixmapItem, Eraser]:
                self.start = None
                return

        if self.status == GraphicsScene.WRITING:
            self.handwriting.clear()

        elif self.status == GraphicsScene.RECTANGLES:
            self.rectangle = self.addRect(QRectF(self.start, self.start), self.color)

        elif self.status == GraphicsScene.ELLIPSES:
            self.rectangle = self.addEllipse(QRectF(self.start, self.start), self.color)

        if self.rectangle is not None:
            self.rectangle.setFlag(QGraphicsItem.ItemIsMovable)
            if self.drawings.get(self.page, None) is None:
                self.drawings[self.page] = []
            self.drawings[self.page].append(self.rectangle)

    def mouseMoveEvent(self, event):
        super().mouseMoveEvent(event)
        if self.status == GraphicsScene.WRITING:
            if self.start is None:
                return

            line = self.addLine(QLineF(self.start, event.scenePos()), self.color)
            self.handwriting.append(line)
            self.start = event.scenePos()

        elif self.status == GraphicsScene.ERASING:
            if self.start is not None:
                self.gum.setPos(event.scenePos() - QLine(50, 50, 50, 50).p2())
                colliding = self.gum.collidingItems()
                for item in colliding:
                    if type(item) in [QGraphicsPixmapItem, QGraphicsProxyWidget]:
                        continue
                    if item in self.drawings.get(self.page, []):
                        self.drawings[self.page].remove(item)
                        self.removeItem(item)

        elif self.status in [GraphicsScene.RECTANGLES, GraphicsScene.ELLIPSES]:
            if self.start is None:
                return
            self.rectangle.setRect(QRectF(self.start, event.scenePos()))

    def mouseReleaseEvent(self, event):
        super().mouseReleaseEvent(event)
        self.start = None
        self.rectangle = None

        if self.status == GraphicsScene.WRITING:
            self.smooth_handwriting()

    def smooth_handwriting(self):
        points = []
        for p in self.handwriting:  # type: QGraphicsLineItem
            points.append((p.line().x1(), p.line().y1()))

        for line in self.handwriting:
            self.removeItem(line)

        smoothed = smooth_with_savgol(points, 20, 4)

        if len(smoothed) > 0:
            p1 = smoothed[0]
            for p2 in smoothed[1:]:
                line = self.addLine(QLineF(QPointF(*p1), QPointF(*p2)), self.color)
                if self.drawings.get(self.page) is None:
                    self.drawings[self.page] = []
                self.drawings[self.page].append(line)
                p1 = p2

    def erase_all(self):
        for item in self.drawings.get(self.page, []):
            self.removeItem(item)
        self.drawings[self.page] = []


class GraphicsView(QGraphicsView):
    resized = pyqtSignal()

    def __init__(self, a):
        super().__init__(a)
        self.setRenderHint(QPainter.SmoothPixmapTransform)
        self.setRenderHint(QPainter.Antialiasing)
        self.setRenderHint(QPainter.TextAntialiasing)
        self.setRenderHint(QPainter.HighQualityAntialiasing)
        # self.scence = None

    def resizeEvent(self, event):
        super().resizeEvent(event)
        transf = QTransform()
        ratio1 = self.viewport().size().width() / self.scene().image.width()
        ratio2 = self.viewport().size().height() / self.scene().image.height()
        ratio = min(ratio1, ratio2)
        transf.scale(ratio, ratio)
        self.setTransform(transf)
        self.scene().setSceneRect(0, 0, self.scene().image.width(), self.scene().image.height())
        self.resized.emit()

        # if self.scence is not None:
        #    self.scene().removeItem(self.scence)
        # self.scence = self.scene().addRect(self.sceneRect(), QColor(255, 0, 0))

    def set_image(self, image, page):
        self.scene().set_image(image, page)


class Slides(QWidget):
    play_code = pyqtSignal(str)

    def set_writing_mode(self, mode):
        for i, elem in enumerate(self.group):
            elem.blockSignals(True)
            elem.setChecked(i == mode)
            elem.blockSignals(False)

        self.scene.status = mode
        self.scene.gum.setVisible(mode == GraphicsScene.ERASING)

        if mode == GraphicsScene.POINTER:
            self.view.viewport().setCursor(QCursor(create_cursor_image()))
        else:
            self.view.viewport().setCursor(Qt.ArrowCursor)

    def set_color(self, color):
        for i, elem in enumerate(self.color_group):
            elem.blockSignals(True)
            elem.setChecked(i == color)
            elem.blockSignals(False)

        colors = [Qt.black, Qt.red, Qt.green, Qt.blue, Qt.yellow, Qt.magenta, Qt.cyan, Qt.gray]
        self.scene.color = QPen(colors[color], self.scene.color.width())
        if self.scene.status in [GraphicsScene.NONE, GraphicsScene.ERASING]:
            self.set_writing_mode(GraphicsScene.WRITING)

    def erase_all(self):
        self.scene.erase_all()

    def create_toolbar(self):
        toolbar = QToolBar()
        # toolbar.setMaximumHeight(35)
        toolbar.show()

        none = toolbar.addAction("", lambda: self.set_writing_mode(0))
        none.setIcon(QIcon(":/icons/cursor.svg"))
        none.setCheckable(True)
        none.setChecked(True)

        pointer = toolbar.addAction("Pointer", lambda: self.set_writing_mode(1))
        pointer.setIcon(QIcon(":/icons/origin.svg"))
        pointer.setCheckable(True)

        write = toolbar.addAction("", lambda: self.set_writing_mode(2))
        write.setIcon(QIcon(":/icons/edit.svg"))
        write.setCheckable(True)

        rectangle = toolbar.addAction(QIcon(":/icons/rectangle.svg"), "", lambda: self.set_writing_mode(4))
        rectangle.setCheckable(True)
        circle = toolbar.addAction(QIcon(":/icons/circle.svg"), "", lambda: self.set_writing_mode(5))
        circle.setCheckable(True)

        toolbar.addSeparator()
        erase = toolbar.addAction("", lambda: self.set_writing_mode(3))
        erase.setIcon(QIcon(":/icons/cancel.svg"))
        erase.setCheckable(True)
        toolbar.addSeparator()

        self.group = [none, pointer, write, erase, rectangle, circle]

        erase_all = toolbar.addAction("", lambda: self.erase_all())
        erase_all.setIcon(QIcon(":/icons/bin.svg"))

        toolbar.addSeparator()
        t1 = toolbar.addAction(QIcon(":/icons/minus.svg"), "", lambda: self.set_thickness(0))
        t1.setCheckable(True)
        t1.setChecked(True)
        t2 = toolbar.addAction(QIcon(":/icons/minus_med.svg"), "", lambda: self.set_thickness(1))
        t2.setCheckable(True)
        t3 = toolbar.addAction(QIcon(":/icons/minus_big.svg"), "", lambda: self.set_thickness(2))
        t3.setCheckable(True)

        self.thickness_group = [t1, t2, t3]

        toolbar.addSeparator()
        black = toolbar.addAction("", lambda: self.set_color(0))
        black.setIcon(QIcon(":/icons/black.svg"))
        red = toolbar.addAction("", lambda: self.set_color(1))
        red.setIcon(QIcon(":/icons/red.svg"))
        green = toolbar.addAction("", lambda: self.set_color(2))
        green.setIcon(QIcon(":/icons/green.svg"))
        blue = toolbar.addAction("", lambda: self.set_color(3))
        blue.setIcon(QIcon(":/icons/blue.svg"))

        self.color_group = [black, red, green, blue]
        for elem in self.color_group:
            elem.setCheckable(True)
        black.setChecked(True)

        toolbar.addSeparator()

        prev = toolbar.addAction("✕", lambda: self.move_to(False))
        prev.setIcon(QIcon(":/icons/arrow-left.svg"))

        self.action_touchable = toolbar.addAction("Hold")
        self.action_touchable.setCheckable(True)
        self.action_touchable.setIcon(QIcon(":/icons/pan.svg"))

        toolbar.addAction(QIcon(":/icons/plus.svg"), "", self.add_empty_page)

        next1 = toolbar.addAction("⬇", lambda: self.move_to(True))
        next1.setIcon(QIcon(":/icons/arrow-right.svg"))
        return toolbar

    def add_empty_page(self):
        self.page += 1
        self.pages_number.insert(self.page, None)
        self.update_image()

    def set_thickness(self, thickness):
        for i, elem in enumerate(self.thickness_group):
            elem.blockSignals(True)
            elem.setChecked(i == thickness)
            elem.blockSignals(False)

        self.scene.color = QPen(self.scene.color.color(), 2 + thickness * 4)
        if self.scene.status in [GraphicsScene.NONE, GraphicsScene.ERASING]:
            self.set_writing_mode(GraphicsScene.WRITING)

    def __init__(self, config, pdf_path, page):
        super().__init__()
        self.config = config
        self.thickness = 2
        self.action_touchable = None
        self.group = []
        self.color_group = []
        self.touchable = True
        self.program = ""
        self.code_buttons = []
        self.code_line_height = 0
        self.resized_pixmap = None
        self.doc = fitz.open(pdf_path)
        self.pages_number = [i for i in range(len(self.doc))]

        self.filename = pdf_path
        self.page = page
        self.base = None

        # Create a QLabel to display the image
        self.scene = GraphicsScene()
        self.scene.navigate.connect(self.navigate)
        self.view = GraphicsView(self.scene)
        self.view.resized.connect(self.resized)
        self.view.setAlignment(Qt.AlignCenter)
        self.view.setSizePolicy(QSizePolicy.Ignored, QSizePolicy.Ignored)

        self.base = QGraphicsRectItem(0, 0, 35, 15, self.scene.pixmap)
        self.base.setBrush(QColor(220, 220, 220))
        self.base.setPen(Qt.transparent)
        self.proxy = QGraphicsProxyWidget(self.base)

        # add shortcut ctrl+n to number of page
        def get_number_of_page():
            a, b = QInputDialog.getInt(self, "Number of page", "Enter the number of page", self.page + 1, 1,
                                       len(self.doc), 1, Qt.WindowFlags())
            self.page = a - 1
            self.update_image()

        q = QShortcut("Ctrl+N", self)
        q.activated.connect(get_number_of_page)

        # Set up a layout
        layout = QHBoxLayout()
        layout.setSpacing(0)
        layout.addWidget(self.view)

        self.setLayout(layout)
        self.update_image()

        self.toolbar_float = None
        self.toolbar = self.create_toolbar()
        self.set_toolbar_float(True)

        QTimer.singleShot(0, self.resize_image)

        QApplication.instance().setStyleSheet("""
                    QToolTip {
                        font-family: 'Courier New', monospace;
                        font-size: 12pt;
                        color: #ffffff;
                        background-color: #333333;
                        border: 1px solid #ffffff;
                    }
                """)

    def is_toolbar_float(self):
        return self.toolbar_float

    def set_toolbar_float(self, value, parent=None):

        if value:
            self.toolbar.setParent(None)
            self.toolbar.setOrientation(Qt.Horizontal)
            self.proxy.setWidget(self.toolbar)
            self.proxy.setPos(-2, 10)
            self.proxy.setZValue(1)
            self.base.setFlags(QGraphicsItem.ItemIgnoresTransformations | QGraphicsItem.ItemIsMovable)
            self.proxy.show()
            self.base.show()
            self.toolbar.show()

        else:
            self.proxy.setWidget(None)
            self.base.hide()
            self.toolbar.setParent(parent)
            self.toolbar.setOrientation(Qt.Horizontal)
            self.toolbar.show()

        self.toolbar_float = value

    def resized(self):
        if self.base is not None:
            self.base.setPos(0, 0)

    def get_toolbar(self):
        return self.toolbar

    def navigate(self, delta):
        self.page = (self.page + delta) % len(self.doc)
        self.update_image()

    def play_program(self, program):
        self.setFocus()
        self.play_code.emit(program)
        return

        # Command to open a new terminal and execute the Python code
        if platform.system() == "Linux":
            subprocess.run([
                "gnome-terminal",
                "--",
                "bash",
                "-c",
                f"python3 -c \"{self.program}\"; exec bash"
            ])
        else:
            # Command to open a new PowerShell window and execute Python code
            subprocess.run([
                "powershell",
                "-NoExit",
                "-Command",
                f"python -c \"{self.program}\""
            ])

    def mousePressEvent(self, a0):
        super().mousePressEvent(a0)
        if a0.button() == Qt.RightButton:
            return

        if self.scene.status != GraphicsScene.NONE:
            return

        if not self.action_touchable.isChecked():
            right_side = a0.x() > self.view.width() // 2
            self.move_to(right_side)

    def move_to(self, right_side):
        if right_side:
            self.page = (self.page + 1) % len(self.doc)
        else:
            self.page = (self.page - 1) % len(self.doc)

        self.update_image()

    def toggle_cursor(self):
        if self.view.cursor() == Qt.ArrowCursor:
            self.set_custom_cursor(self.view)
        else:
            self.view.setCursor(Qt.ArrowCursor)
        self.view.update()

    def update_image(self):
        # Clear buttons
        for button, _ in self.code_buttons:
            self.scene.removeItem(button)
        self.code_buttons.clear()

        # Get number of page
        pdf_page = self.pages_number[self.page]
        if pdf_page is None:
            image = QPixmap(1920, 1080)
            image.fill(Qt.white)
        else:
            page = self.doc[pdf_page]
            pix = page.get_pixmap(matrix=fitz.Matrix(2, 2), alpha=False, annots=True)
            image = QImage(pix.samples, pix.width, pix.height, pix.stride, QImage.Format_RGB888)
            for d in page.get_drawings():
                fill = d.get("fill")
                type = d.get("type")
                color = d.get("color")

                if type != 'f' and color == (1.0, 0, 1.0):
                    rect = d.get("rect")
                    self.extract_text_and_fonts(rect)

            self.update_button_pos()

        self.pixmap = QPixmap(image)
        self.resize_image()
        self.view.set_image(self.pixmap, self.page)

    def resizeEvent(self, event):
        super().resizeEvent(event)
        self.resize_image()

    def resize_image(self):
        # Resize the pixmap to fit the QLabel while maintaining aspect ratio
        if not self.pixmap is None and not self.pixmap.isNull():
            self.resized_pixmap = self.pixmap.scaled(
                self.view.size(),
                Qt.KeepAspectRatio,
                Qt.SmoothTransformation
            )

    def extract_text_and_fonts(self, rect=None):
        lines = []
        try:

            page_without_empty = 0
            for i in range(self.page+1):
                page_without_empty += 1 if self.pages_number[i] is not None else 0

            page = self.doc[page_without_empty - 1]

            # Extract blocks of text
            blocks = page.get_text("dict", clip=rect)['blocks']
            for block in blocks:
                if 'lines' in block:  # Ensure the block contains text

                    block_bbox = block['bbox']  # Get block position
                    for line in block['lines']:
                        line_text = ""
                        line_box = line["bbox"]
                        for span in line['spans']:
                            text = span['text']
                            font = span['font']
                            position = span['bbox']  # Get position of the span
                            line_text += text

                        lines.append((line_text, line_box))

        except Exception as e:
            print(f"An error occurred: {e}")

        if len(lines) != 0:
            rect: Rect

            xs = set()
            for text, pos in lines:
                x1, y1, x2, y2 = pos
                x1 = int(x1)
                xs.add(x1)
            xs = list(xs)
            xs.sort()

            program = str()
            for text, pos in lines:
                x1, y1, x2, y2 = pos
                x1 = int(x1)
                i = xs.index(x1) * 4
                program += " " * i + text + "\n"

            code_pos = (rect.x0, rect.y0, rect.x1 - rect.x0, rect.y1 - rect.y0)

            play_button = QPushButton()
            play_button.setFixedSize(45, 45)
            play_button.setIcon(self.style().standardIcon(QApplication.style().SP_MediaPlay))
            play_button.clicked.connect(lambda x=program, y=program: self.play_program(y))

            proxy = QGraphicsProxyWidget(self.scene.pixmap)
            proxy.setWidget(play_button)
            # proxy.setFlags(QGraphicsItem.ItemIgnoresTransformations)

            # Add the proxy widget to the scene
            # self.scene.addItem(proxy)

            self.code_buttons.append((proxy, code_pos))
            play_button.setToolTip(self.program)

    def update_button_pos(self):

        for button, code_pos in self.code_buttons:
            code_x, code_y, code_w, code_h = code_pos
            button: QGraphicsRectItem
            button.setPos((code_x + code_w) * 2 - button.sceneBoundingRect().width(),
                          (code_y + code_h) * 2 - button.sceneBoundingRect().height())



from PyQt5.QtCore import Qt, QRegExp
from PyQt5.QtGui import QSyntaxHighlighter, QTextCharFormat, QColor, QFont


class SyntaxHighlighter(QSyntaxHighlighter):

    def __init__(self, keywords):
        super().__init__(None)
        self.highlighting_rules = []
        self.keywords = keywords
        self.dark = False
        self.apply_scheme()

    def set_dark_mode(self, dark):
        self.dark = dark

        self.highlighting_rules.clear()
        self.apply_scheme()

    def apply_scheme(self):
        for keywords, color1, color2 in self.keywords:
            keyword_format = QTextCharFormat()
            keyword_format.setForeground(color1 if not self.dark else color2)
            keyword_format.setFontWeight(QFont.Bold)

            self.highlighting_rules += [(f"\\b{k}\\b", keyword_format) for k in keywords]

        string_format = QTextCharFormat()
        string_format.setForeground(Qt.magenta)
        self.highlighting_rules.append((r'".*"', string_format))
        self.highlighting_rules.append((r"'.*'", string_format))

        comment_format = QTextCharFormat()
        comment_format.setForeground(QColor("green"))
        comment_format.setFontItalic(True)
        self.highlighting_rules.append((r"#.*", comment_format))

    def highlightBlock(self, text):
        for pattern, fmt in self.highlighting_rules:
            expression = QRegExp(pattern)
            index = expression.indexIn(text)
            while index >= 0:
                length = expression.matchedLength()
                self.setFormat(index, length, fmt)
                index = expression.indexIn(text, index + length)

    def get_keywords(self):
        return self.keywords


class PythonHighlighter(SyntaxHighlighter):
    def __init__(self, dark=False):
        super().__init__([
            (['return', 'nonlocal', 'elif', 'assert', 'or', 'yield', 'finally',
              'from', 'global', 'del', 'print', 'None', 'pass', 'class', 'as',
              'break', 'while', 'await', 'async', 'range', 'is', 'True', 'lambda',
              'False', 'in', 'import', 'except', 'continue', 'and', 'raise', 'with',
              'if', 'try', 'for', 'else', 'not', 'def', "input", "int", "float", "str",
              "list", "dict", "input", "print", "open", "read", "write", "close", "split",
              ], Qt.blue, Qt.cyan),
            (["self"], Qt.darkCyan, Qt.darkCyan)
        ])


class PascalHighlighter(SyntaxHighlighter):
    def __init__(self, dark=False):
        super().__init__([([
                               "and", "array", "asm", "begin", "case", "const", "constructor", "destructor",
                               "div", "do", "downto", "else", "end", "file", "for", "function", "goto", "if",
                               "implementation", "in", "inherited", "inline", "interface", "label", "mod", "nil",
                               "not", "object", "of", "or", "packed", "procedure", "program", "record", "repeat",
                               "set", "shl", "shr", "string", "then", "to", "type", "unit", "until", "uses",
                               "var", "while", "with", "xor", "AND", "ARRAY", "ASM", "BEGIN", "CASE", "CONST", "CONSTRUCTOR", "DESTRUCTOR",
                               "DIV", "DO", "DOWNTO", "ELSE", "END", "FILE", "FOR", "FUNCTION", "GOTO", "IF",
                               "IMPLEMENTATION", "IN", "INHERITED", "INLINE", "INTERFACE", "LABEL", "MOD", "NIL",
                               "NOT", "OBJECT", "OF", "OR", "PACKED", "PROCEDURE", "PROGRAM", "RECORD", "REPEAT",
                               "SET", "SHL", "SHR", "STRING", "THEN", "TO", "TYPE", "UNIT", "UNTIL", "USES",
                               "VAR", "WHILE", "WITH", "XOR"
                           ], Qt.blue, Qt.blue)])



from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QPushButton, QTextEdit, QVBoxLayout, QDialog



class Author(QDialog):
    def __init__(self):
        super().__init__()
        self.setLayout(QVBoxLayout())
        self.setMinimumSize(400, 350)
        self.setWindowTitle("About")
        textEdit = QTextEdit()
        textEdit.setReadOnly(True)
        textEdit.setHtml("""<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <title>SPICE - About</title>
    <style>
        body {
            font-family: 'Arial', sans-serif;
            background-color: #f4f4f9;
            color: #333;
            margin: 0;
            padding: 0;
            display: flex;
            justify-content: center;
            align-items: center;
            height: 100vh;
            text-align: center;
        }

        .container {
            background-color: #ffffff;
            padding: 20px;
            border-radius: 8px;
            box-shadow: 0 4px 8px rgba(0, 0, 0, 0.1);
            width: 80%;
            max-width: 600px;
        }

        h1 {
            font-size: 3em;
            color: #2c3e50;
            margin-bottom: 20px;
        }

        p {
            font-size: 1.2em;
            line-height: 1.6;
        }

        ul {
            list-style: none;
            padding: 0;
        }

        ul li {
            font-size: 1.1em;
            margin: 5px 0;
        }

        .footer {
            font-size: 1em;
            color: #7f8c8d;
            margin-top: 20px;
        }

        .footer p {
            margin: 0;
        }

    </style>
</head>
<body>

    <div class="container">
        <h1>Spice</h1>
        <h2><strong>Slides and Python for Interactive and Creative Education</strong></h2>

        <h4><strong>Developed by:</strong></h4>
            <h2>Danilo Tardioli</h2>
            <h3>Email: <a href="mailto:dantard@unizar.es">dantard@unizar.es</a></h3>

        <p><strong>Year:</strong> 2024</p>

        <div class="footer">
            <p><strong>Learn more at:</strong> <a href="https://github.com/dantard/coder">https://github.com/dantard/coder</a></p>
        </div>
    </div>

</body>
</html>

        """)
        self.layout().addWidget(textEdit)
        close_button = QPushButton("Close")
        close_button.setMaximumWidth(100)
        # center the button
        self.layout().addWidget(close_button)
        self.layout().setAlignment(close_button, Qt.AlignRight)
        close_button.clicked.connect(self.close)


import os
import shutil

from PyQt5 import QtGui
from PyQt5.QtCore import QObject, pyqtSignal, QDir, QItemSelectionModel, QModelIndex, Qt, QTimer
from PyQt5.QtWidgets import QWidget, QTreeView, QFileSystemModel, QVBoxLayout, QPushButton, QHBoxLayout, QLabel, QMenu, QMessageBox, QToolBar, QInputDialog


class Tree(QTreeView):
    delete_requested = pyqtSignal(str)

    def filter_rows(self, extensions=None):

        extensions = extensions or [".txt", ".py", ".csv"]

        for i in range(self.model().rowCount(self.rootIndex())):
            child_index = self.model().index(i, 0, self.rootIndex())  # Get index of each row
            filename = self.model().data(child_index)  # Get file name
            full_path = self.model().filePath(child_index)
            if os.path.isdir(full_path):
                if "__pycache__" in full_path:
                    self.setRowHidden(i, self.rootIndex(), True)
            else:
                extension = os.path.splitext(filename)[1]
                if extension not in extensions:
                    self.setRowHidden(i, self.rootIndex(), True)

    def contextMenuEvent(self, a0: QtGui.QContextMenuEvent) -> None:
        super().contextMenuEvent(a0)
        indexes = self.selectedIndexes()
        if indexes:
            index = self.indexAt(a0.pos())
            if index.isValid():
                dirModel = self.model()
                path = dirModel.fileInfo(index).absoluteFilePath()
                menu = QMenu()
                delete = menu.addAction("Delete")
                res = menu.exec_(self.viewport().mapToGlobal(a0.pos()))
                if res == delete:
                    self.delete_requested.emit(path)


class FileBrowser(QWidget):
    class Signals(QObject):
        file_selected = pyqtSignal(str)

    def __init__(self, path, filters=None, hide_details=True):
        super().__init__()
        self.current_files = []
        self.signals = self.Signals()
        self.path = path
        self.treeview = Tree()
        self.treeview.delete_requested.connect(self.delete_requested)
        self.dirModel = QFileSystemModel()
        self.dirModel.directoryLoaded.connect(lambda:self.treeview.filter_rows(filters))
        #self.dirModel.setNameFilters(filters)
        self.dirModel.setNameFilterDisables(False)

        self.treeview.setModel(self.dirModel)
        self.treeview.setRootIndex(self.dirModel.setRootPath(path))

        vlayout = QVBoxLayout(self)
        vlayout.setSpacing(0)
        vlayout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(vlayout)
        #tb = QToolBar()
        #tb.addAction("🗀", self.refresh)

        #vlayout.addWidget(tb)

        self.layout().addWidget(self.treeview)
        self.treeview.selectionModel().selectionChanged.connect(self.on_current_changed)
        self.treeview.doubleClicked.connect(self.on_double_clicked)

        if hide_details:
            for i in range(1, self.treeview.model().columnCount()):
                self.treeview.header().hideSection(i)

    def delete_requested(self, path):
        if QMessageBox.question(self, "Delete", f"Are you sure you want to delete {path}?", QMessageBox.Yes | QMessageBox.No) == QMessageBox.No:
            return
        if os.path.isdir(path):
            shutil.rmtree(path)
        else:
            os.remove(path)




    def on_double_clicked(self, index):
        # Map the proxy index to the source model index
        path = self.dirModel.fileInfo(index).absoluteFilePath()
        if os.path.isdir(path):
            return
        self.signals.file_selected.emit(path)

    def btn_up_clicked(self):
        index = self.treeview.rootIndex()
        if index.isValid():
            index = index.parent()
            self.set_root_index(index)

    def set_root(self, path):
        if not os.path.exists(path):
            return
        self.treeview.setRootIndex(self.dirModel.setRootPath(path))
        self.current_files = [str(x) for  x in os.listdir(path)]

    def set_root_index(self, index):
        self.treeview.setRootIndex(index)
        path = self.dirModel.fileInfo(index).absoluteFilePath()
        self.label.setText(path)

    def select(self, filename, emit=True):
        if not emit:
            self.treeview.selectionModel().blockSignals(True)
        index = self.dirModel.index(filename)
        indices = []
        while index.isValid():
            indices.append(index)
            index = index.parent()

        for index in reversed(indices):
            self.treeview.setExpanded(index, True)

        self.treeview.setCurrentIndex(index)
        self.treeview.selectionModel().blockSignals(False)

    def on_current_changed(self, selected, deselected):
        pass
        # if deselected.indexes():
        #     print("deselected1", deselected.indexes())
        #     # Check if the deselected index is valid
        #     for index in deselected.indexes():
        #         if not os.path.isfile(self.dirModel.filePath(index)):
        #             self.treeview.clearSelection()
        #             return
        #
        # if selected is None or len(selected.indexes()) < 1:
        #     return
        # path = self.dirModel.fileInfo(selected.indexes()[0]).absoluteFilePath()
        # # self.listview.setRootIndex(self.fileModel.setRootPath(path))
        # if os.path.isfile(path):
        #     self.signals.file_selected.emit(path)

    def new_folder(self):
        current_path = self.dirModel.rootPath()

        # If the toolbar refresh button was clicked with no arguments,
        # show the folder creation dialog
        folder_name, ok = QInputDialog.getText(self, "Folder Name", "Enter the folder name")
        if ok and folder_name:
            new_folder_path = os.path.join(self.path, folder_name)
            os.makedirs(new_folder_path, exist_ok=True)
            self.dirModel.setRootPath("")
            self.dirModel.setRootPath(current_path)



from PyQt5.QtGui import QPainter
from PyQt5.QtWidgets import QScrollBar


class MagicScrollBar(QScrollBar):
    def paintEvent(self, a0) -> None:
        super().paintEvent(a0)
        if self.maximum() == 0:
            p = QPainter(self)
            p.fillRect(self.rect(), self.parent().palette().base().color())

